#==================================================================
# select_testcase.py
# This Python script loads the test case file from the library
# ---Author---
# W. Bukhsh,
# wbukhsh@gmail.com
# OATS
# Copyright (c) 2015 by W Bukhsh, Glasgow, Scotland
# OATS is distributed under the GNU GENERAL PUBLIC LICENSE v3 (see LICENSE file for details).
#==================================================================

import pandas as pd

def selecttestcase(test):
    xl = pd.ExcelFile(test)

    df_bus         = xl.parse("bus")
    df_demand      = xl.parse("demand")
    df_branch      = xl.parse("branch")
    df_generators  = xl.parse("generator")
    df_shunt       = xl.parse("shunt")
    df_transformer = xl.parse("transformer")
    df_wind        = xl.parse("wind")
    df_baseMVA     = xl.parse("baseMVA")
    df_zone        = xl.parse("zone")
    df_zonalNTC    = xl.parse("zonalNTC")
    df_ts          = xl.parse("timeseries",header=[0,1])


    data = {
    "bus": df_bus,
    "demand": df_demand,
    "branch": df_branch,
    "generator": df_generators,
    "shunt": df_shunt,
    "transformer": df_transformer,
    "wind": df_wind,
    "baseMVA": df_baseMVA,
    "zone": df_zone,
    "timeseries": df_ts,
    "zone":df_zone,
    "zonalNTC":df_zonalNTC
    }
    try:
        df_storage   = xl.parse("storage")
        data['storage'] = df_storage
    except:
        print ('Storage not defined')

    return data



#==================================================================
# printout.py
# A Python script to write output to xls and on screen
# ---Author---
# W. Bukhsh,
# wbukhsh@gmail.com
# OATS
# Copyright (c) 2015 by W Bukhsh, Glasgow, Scotland
# OATS is distributed under the GNU GENERAL PUBLIC LICENSE v3 (see LICENSE file for details).
#==================================================================
from pyomo.opt import SolverStatus, TerminationCondition
from tabulate import tabulate
import pandas as pd
import math
import sys
class printoutput(object):
    def __init__(self, results, instance,mod):
        self.results   = results
        self.instance  = instance
        self.mod       = mod
    def greet(self):
        print ("========================")
        print ("\n Output from the OATS")
        print ("========================")
    def solutionstatus(self):
        self.instance.solutions.load_from(self.results)
        print ("------Solver Message------")
        print (self.results.solver)
        print ("--------------------------")
        if (self.results.solver.status == SolverStatus.ok) \
        and (self.results.solver.termination_condition == TerminationCondition.optimal):
            print ("Optimization Converged!")
        elif self.results.solver.termination_condition == TerminationCondition.infeasible:
            sys.exit("Problem is infeasible!\nOats terminated. No output is written on the results file.")
        else:
            print (sys.exit("Problem is infeasible!\nOats terminated. No output is written on the results file."))
    def printsummary(self):
        if 'LF' not in self.mod:
            print ("Cost of the objective function:", str(float(self.instance.OBJ())))
        print ("***********")
        print ("\n Summary")
        print ("***********")
        tab_summary = []
        tab_summary.append(['Conventional generation (MW)','Wind generation (MW)', 'Demand (MW)'])
        tab_summary.append([sum(self.instance.pG[g].value for g in self.instance.G)*self.instance.baseMVA,\
        sum(self.instance.pW[w].value for w in self.instance.WIND)*self.instance.baseMVA,sum(self.instance.PD[d] for d in self.instance.D)*self.instance.baseMVA])
        print (tabulate(tab_summary, headers="firstrow", tablefmt="grid"))
        print ("==============================================")
    def printoutputxls(self):
        #===initialise pandas dataframes
        cols_summary    = ['Conventional generation (MW)', 'Wind generation (MW)', 'Demand (MW)','Objective function value']
        cols_bus        = ['name', 'angle(degs)']
        cols_demand     = ['name', 'busname', 'PD(MW)']

        if ('DC' in self.mod) or ('SC' in self.mod):
            cols_branch     = ['name', 'from_busname', 'to_busname', 'pL(MW)']
            cols_transf     = ['name', 'from_busname', 'to_busname', 'pLT(MW)']
            if 'LF' in self.mod:
                cols_generation = ['name', 'busname', 'PG(MW)', 'pG(MW)']
                cols_wind       = ['name', 'busname', 'PG(MW)', 'pG(MW)']
            elif 'OPF' in self.mod:
                cols_demand.append('alpha')
                cols_branch.append('SLmax(MW)')
                cols_transf.append('SLmax(MW)')
                cols_generation = ['name', 'busname', 'PGLB(MW)','PG(MW)', 'pG(MW)','PGUB(MW)']
                cols_wind       = ['name', 'busname', 'PGLB(MW)','PG(MW)', 'pG(MW)','PGUB(MW)']
        elif 'AC' in self.mod:
            cols_bus.append('Voltage(p.u.)')
            cols_demand.append('QD(MVar)')
            cols_branch     = ['name', 'from_busname', 'to_busname', 'pLto(MW)', 'pLfrom(MW)', 'loss(MW)']
            cols_transf     = ['name', 'from_busname', 'to_busname', 'pLTto(MW)', 'pLTfrom(MW)', 'loss(MW)']
            if 'LF' in self.mod:
                cols_transf.append('tap')
                cols_generation = ['name', 'busname', 'PG(MW)', 'pG(MW)', 'qG(MVar)']
                cols_wind       = ['name', 'busname', 'PG(MW)', 'pG(MW)', 'qG(MVar)']
            elif 'OPF' in self.mod:
                cols_demand.append('alpha')
                cols_generation = ['name', 'busname', 'PGLB(MW)', 'PG(MW)', 'pG(MW)','PGUB(MW)', 'QGLB(MVar)', 'qG(MVar)','QGUB(MVar)']
                cols_wind       = ['name', 'busname', 'PGLB(MW)', 'PG(MW)', 'pG(MW)','PGUB(MW)', 'QGLB(MVar)', 'qG(MVar)','QGUB(MVar)']

        summary         = pd.DataFrame(columns=cols_summary)
        bus             = pd.DataFrame(columns=cols_bus)
        demand          = pd.DataFrame(columns=cols_demand)
        wind            = pd.DataFrame(columns=cols_wind)
        generation      = pd.DataFrame(columns=cols_generation)
        branch          = pd.DataFrame(columns=cols_branch)
        transformer     = pd.DataFrame(columns=cols_transf)

        #-----write Data Frames

        summary.loc[0] = pd.Series({'Conventional generation (MW)': sum(self.instance.pG[g].value for g in self.instance.G)*self.instance.baseMVA,\
        'Wind generation (MW)':sum(self.instance.pW[w].value for w in self.instance.WIND)*self.instance.baseMVA,\
        'Demand (MW)':sum(self.instance.PD[d] for d in self.instance.D)*self.instance.baseMVA,\
        'Objective function value': self.instance.OBJ()})

        if ('DC' in self.mod) or ('SC' in self.mod):
            #bus data
            ind=0
            for b in self.instance.B:
                bus.loc[ind] = pd.Series({'name': b,'angle(degs)':self.instance.delta[b].value*180/math.pi})
                ind += 1
            if 'LF' in self.mod:
                #line data
                ind=0
                for b in self.instance.L:
                    branch.loc[ind] = pd.Series({'name': b, 'from_busname':self.instance.A[b,1], 'to_busname':self.instance.A[b,2],\
                    'pL(MW)':self.instance.pL[b].value*self.instance.baseMVA})
                    ind += 1
                #transformer data
                ind = 0
                for b in self.instance.TRANSF:
                    transformer.loc[ind] = pd.Series({'name': b, 'from_busname':self.instance.AT[b,1],
                    'to_busname':self.instance.AT[b,2], 'pLT(MW)':self.instance.pLT[b].value*self.instance.baseMVA})
                ind += 1
                #demand data
                ind = 0
                for d in self.instance.Dbs:
                    demand.loc[ind] = pd.Series({'name': d[1],'busname':d[0],'PD(MW)':self.instance.PD[d[1]]*self.instance.baseMVA})
                    ind += 1
                #generator data
                ind = 0
                for g in self.instance.Gbs:
                    generation.loc[ind] = pd.Series({'name': g[0], 'busname':g[1],\
                    'PG(MW)':self.instance.PG[g[1]]*self.instance.baseMVA,'pG(MW)':round(self.instance.pG[g[1]].value*self.instance.baseMVA,3)})
                    ind += 1
                ind = 0
                #wind data
                for w in self.instance.Wbs:
                    wind.loc[ind] = pd.Series({'name': w[0], 'busname':w[1],\
                    'PG(MW)':self.instance.WG[w[1]]*self.instance.baseMVA, 'pG(MW)':round(self.instance.pW[w[1]].value*self.instance.baseMVA,3)})
                    ind += 1
            elif 'BM' in self.mod:
                #demand data
                ind = 0
                for d in self.instance.Dbs:
                    demand.loc[ind] = pd.Series({'name': d[1],'busname':d[0],'PD(MW)':self.instance.PD[d[1]]*self.instance.baseMVA,\
                    'alpha':round(self.instance.alpha[d[1]].value,3)})
                    ind += 1
                #generator data
                ind = 0
                for g in self.instance.Gbs:
                    generation.loc[ind] = pd.Series({'name':g[0], 'busname':g[1],\
                    'PGLB(MW)':self.instance.PGmin[g[1]]*self.instance.baseMVA,\
                    'PG(MW)':round(self.instance.PG[g[1]]*self.instance.baseMVA,3),\
                    'pG(MW)':round(self.instance.pG[g[1]].value*self.instance.baseMVA,3),\
                    'PGUB(MW)':self.instance.PGmax[g[1]]*self.instance.baseMVA})
                    ind += 1
                #wind data
                ind = 0
                for g in self.instance.Wbs:
                    wind.loc[ind] = pd.Series({'name':g[0], 'busname':g[1],\
                    'PGLB(MW)':self.instance.WGmin[g[1]]*self.instance.baseMVA,\
                    'PG(MW)':round(self.instance.PW[g[1]]*self.instance.baseMVA,3),\
                    'pG(MW)':round(self.instance.pW[g[1]].value*self.instance.baseMVA,3),\
                    'PGUB(MW)':self.instance.WGmax[g[1]]*self.instance.baseMVA})
                    ind += 1

            elif 'OPF' in self.mod:
                #line data
                ind=0
                for b in self.instance.L:
                    branch.loc[ind] = pd.Series({'name': b, 'from_busname':self.instance.A[b,1], 'to_busname':self.instance.A[b,2],\
                    'pL(MW)':self.instance.pL[b].value*self.instance.baseMVA,'SLmax(MW)':self.instance.SLmax[b]*self.instance.baseMVA})
                    ind += 1
                #transformer data
                ind = 0
                for b in self.instance.TRANSF:
                    transformer.loc[ind] = pd.Series({'name': b, 'from_busname':self.instance.AT[b,1],
                    'to_busname':self.instance.AT[b,2], 'pLT(MW)':self.instance.pLT[b].value*self.instance.baseMVA,\
                    'SLmax(MW)':self.instance.SLmaxT[b]*self.instance.baseMVA})
                    ind += 1

                #demand data
                ind = 0
                for d in self.instance.Dbs:
                    demand.loc[ind] = pd.Series({'name': d[1],'busname':d[0],'PD(MW)':self.instance.PD[d[1]]*self.instance.baseMVA,\
                    'alpha':round(self.instance.alpha[d[1]].value,3)})
                    ind += 1
                #generator data
                ind = 0
                for g in self.instance.Gbs:
                    generation.loc[ind] = pd.Series({'name':g[0], 'busname':g[1],\
                    'PGLB(MW)':self.instance.PGmin[g[1]]*self.instance.baseMVA,\
                    'pG(MW)':round(self.instance.pG[g[1]].value*self.instance.baseMVA,3),\
                    'PGUB(MW)':self.instance.PGmax[g[1]]*self.instance.baseMVA})
                    ind += 1

                #wind data
                ind = 0
                for g in self.instance.Wbs:
                    wind.loc[ind] = pd.Series({'name':g[0], 'busname':g[1],\
                    'PGLB(MW)':self.instance.WGmin[g[1]]*self.instance.baseMVA,\
                    'PG(MW)':round(self.instance.PW[g[1]]*self.instance.baseMVA,3),\
                    'pG(MW)':round(self.instance.pW[g[1]].value*self.instance.baseMVA,3),\
                    'PGUB(MW)':self.instance.WGmax[g[1]]*self.instance.baseMVA})
                    ind += 1
        elif 'AC' in self.mod:
            #bus data
            ind=0
            for b in self.instance.B:
                bus.loc[ind] = pd.Series({'name': b,'angle(degs)':self.instance.delta[b].value*180/math.pi,\
                'Voltage(p.u.)':self.instance.v[b].value})
                ind += 1
            #line data
            ind = 0
            for l in self.instance.L:
                branch.loc[ind] = pd.Series({'name': l, 'from_busname':self.instance.A[l,1], 'to_busname':self.instance.A[l,2],\
                'pLto(MW)':self.instance.pLto[l].value*self.instance.baseMVA,\
                'pLfrom(MW)':self.instance.pLfrom[l].value*self.instance.baseMVA,\
                'loss(MW)':(self.instance.pLto[l].value+self.instance.pLfrom[l].value)*self.instance.baseMVA})
                ind += 1
            ind = 0
            if 'LF' in self.mod:
                #transformer data
                for l in self.instance.TRANSF:
                    transformer.loc[ind] = pd.Series({'name': l, 'from_busname':self.instance.AT[l,1],
                    'to_busname':self.instance.AT[l,2], 'pLTto(MW)':self.instance.pLtoT[l].value*self.instance.baseMVA,\
                    'pLTfrom(MW)':self.instance.pLfromT[l].value*self.instance.baseMVA,\
                    'loss(MW)':(self.instance.pLfromT[l].value+self.instance.pLtoT[l].value)*self.instance.baseMVA,\
                    'tap':self.instance.tap[l].value})
                    ind += 1
                #demand data
                ind = 0
                for d in self.instance.Dbs:
                    demand.loc[ind] = pd.Series({'name': d[1],'busname':d[0],'PD(MW)':self.instance.PD[d[1]]*self.instance.baseMVA,\
                    'QD(MVar)':self.instance.QD[d[1]]*self.instance.baseMVA})
                    ind += 1
                #generator data
                ind = 0
                for g in self.instance.Gbs:
                    generation.loc[ind] = pd.Series({'name': g[0], 'busname':g[1],\
                    'PG(MW)':self.instance.PG[g[1]]*self.instance.baseMVA,'pG(MW)':round(self.instance.pG[g[1]].value*self.instance.baseMVA,3),\
                    'qG(MVar)':round(self.instance.qG[g[1]].value*self.instance.baseMVA,3)})
                    ind += 1
                #wind data
                ind = 0
                for g in self.instance.Wbs:
                    wind.loc[ind] = pd.Series({'name':g[0], 'busname':g[1],\
                    'PG(MW)':self.instance.PG[g[1]]*self.instance.baseMVA,\
                    'pG(MW)':round(self.instance.pW[g[1]].value*self.instance.baseMVA,3)})
                    ind += 1
            elif 'BM' in self.mod:
                ind = 0
                #transformer data
                for l in self.instance.TRANSF:
                    transformer.loc[ind] = pd.Series({'name': l, 'from_busname':self.instance.AT[l,1],
                    'to_busname':self.instance.AT[l,2], 'pLTto(MW)':self.instance.pLtoT[l].value*self.instance.baseMVA,\
                    'pLTfrom(MW)':self.instance.pLfromT[l].value*self.instance.baseMVA,\
                    'loss(MW)':(self.instance.pLfromT[l].value+self.instance.pLtoT[l].value)*self.instance.baseMVA})
                    ind += 1
                #demand data
                ind = 0
                for d in self.instance.Dbs:
                    demand.loc[ind] = pd.Series({'name': d[1],'busname':d[0],'PD(MW)':self.instance.PD[d[1]]*self.instance.baseMVA,\
                    'QD(MVar)':self.instance.QD[d[1]]*self.instance.baseMVA,\
                    'alpha':round(self.instance.alpha[d[1]].value,3)})
                    ind += 1
                #generator data
                ind=0
                for g in self.instance.Gbs:
                    generation.loc[ind] = pd.Series({'name':g[0], 'busname':g[1],\
                    'PGLB(MW)':self.instance.PGmin[g[1]]*self.instance.baseMVA,\
                    'PG(MW)':round(self.instance.PG[g[1]]*self.instance.baseMVA,3),\
                    'pG(MW)':round(self.instance.pG[g[1]].value*self.instance.baseMVA,3),\
                    'PGUB(MW)':self.instance.PGmax[g[1]]*self.instance.baseMVA,\
                    'QGLB(MVar)':self.instance.QGmin[g[1]]*self.instance.baseMVA,\
                    'qG(MVar)':round(self.instance.qG[g[1]].value*self.instance.baseMVA,3),\
                    'QGUB(MVar)':self.instance.QGmax[g[1]]*self.instance.baseMVA})
                    ind += 1
                #wind data
                ind = 0
                for g in self.instance.Wbs:
                    wind.loc[ind] = pd.Series({'name':g[0], 'busname':g[1],\
                    'PGLB(MW)':self.instance.PGmin[g[1]]*self.instance.baseMVA,\
                    'pG(MW)':round(self.instance.pW[g[1]].value*self.instance.baseMVA,3),\
                    'PGUB(MW)':self.instance.PGmax[g[1]]*self.instance.baseMVA,\
                    'QGLB(MVar)':self.instance.QGmin[g[1]]*self.instance.baseMVA,\
                    'qG(MVar)':round(self.instance.qW[g[1]].value*self.instance.baseMVA,3),\
                    'QGUB(MVar)':self.instance.QGmax[g[1]]*self.instance.baseMVA})
                    ind += 1
            elif 'OPF' in self.mod:
                ind = 0
                #transformer data
                for l in self.instance.TRANSF:
                    transformer.loc[ind] = pd.Series({'name': l, 'from_busname':self.instance.AT[l,1],
                    'to_busname':self.instance.AT[l,2], 'pLTto(MW)':self.instance.pLtoT[l].value*self.instance.baseMVA,\
                    'pLTfrom(MW)':self.instance.pLfromT[l].value*self.instance.baseMVA,\
                    'loss(MW)':(self.instance.pLfromT[l].value+self.instance.pLtoT[l].value)*self.instance.baseMVA})
                    ind += 1
                #demand data
                ind = 0
                for d in self.instance.Dbs:
                    demand.loc[ind] = pd.Series({'name': d[1],'busname':d[0],'PD(MW)':self.instance.PD[d[1]]*self.instance.baseMVA,\
                    'QD(MVar)':self.instance.QD[d[1]]*self.instance.baseMVA,\
                    'alpha':round(self.instance.alpha[d[1]].value,3)})
                    ind += 1
                #generator data
                ind=0
                for g in self.instance.Gbs:
                    generation.loc[ind] = pd.Series({'name':g[0], 'busname':g[1],\
                    'PGLB(MW)':self.instance.PGmin[g[1]]*self.instance.baseMVA,\
                    'pG(MW)':round(self.instance.pG[g[1]].value*self.instance.baseMVA,3),\
                    'PGUB(MW)':self.instance.PGmax[g[1]]*self.instance.baseMVA,\
                    'QGLB(MVar)':self.instance.QGmin[g[1]]*self.instance.baseMVA,\
                    'qG(MVar)':round(self.instance.qG[g[1]].value*self.instance.baseMVA,3),\
                    'QGUB(MVar)':self.instance.QGmax[g[1]]*self.instance.baseMVA})
                    ind += 1
                #wind data
                ind = 0
                for g in self.instance.Wbs:
                    wind.loc[ind] = pd.Series({'name':g[0], 'busname':g[1],\
                    'PGLB(MW)':self.instance.PGmin[g[1]]*self.instance.baseMVA,\
                    'pG(MW)':round(self.instance.pW[g[1]].value*self.instance.baseMVA,3),\
                    'PGUB(MW)':self.instance.PGmax[g[1]]*self.instance.baseMVA,\
                    'QGLB(MVar)':self.instance.QGmin[g[1]]*self.instance.baseMVA,\
                    'qG(MVar)':round(self.instance.qW[g[1]].value*self.instance.baseMVA,3),\
                    'QGUB(MVar)':self.instance.QGmax[g[1]]*self.instance.baseMVA})
                    ind += 1


        #----------------------------------------------------------
        #===write output on xlsx file===
        #
        bus = bus.sort_values(['name'])
        generation = generation.sort_values(['name'])
        demand = demand.sort_values(['name'])
        writer = pd.ExcelWriter('results.xlsx', engine ='xlsxwriter')
        summary.to_excel(writer, sheet_name = 'summary',index=False)
        bus.to_excel(writer, sheet_name = 'bus',index=False)
        demand.to_excel(writer, sheet_name = 'demand',index=False)
        generation.to_excel(writer, sheet_name = 'generator',index=False)
        wind.to_excel(writer, sheet_name = 'wind',index=False)
        branch.to_excel(writer, sheet_name = 'branch',index=False)
        transformer.to_excel(writer, sheet_name = 'transformer',index=False)
        writer.save()
    def printUC(self):
        print ("Cost of the objective function:", str(float(self.instance.OBJ())))
        print ("***********")
        print ("\n Summary")
        print ("***********")
        tab_summary = []
        tab_summary.append(['Time Period','Units on','Conventional generation (MW)','Wind generation (MW)', 'Demand (MW)','Cost'])
        for t in self.instance.T:
            tab_summary.append([t,sum(self.instance.u[g,t].value for g in self.instance.G),\
            sum(self.instance.pG[g,t].value for g in self.instance.G)*self.instance.baseMVA,\
            sum(self.instance.pW[w,t].value for w in self.instance.WIND)*self.instance.baseMVA,sum(self.instance.PD[d,t] for d in self.instance.D)*self.instance.baseMVA,\
            self.instance.costfunc[t].value/(sum(self.instance.PD[d,t] for d in self.instance.D)*self.instance.baseMVA)])
        print (tabulate(tab_summary, headers="firstrow", tablefmt="grid"))

        print ("***************\n")
        print ("Zonal Flows\n")
        print ("***************")
        zoneflow = []
        zoneflow.append(['Time Period','Fr','To','Zone Fr (MW)','Zone To (MW)'])
        for t in self.instance.T:
            for l in self.instance.ICT:
                zoneflow.append([t,self.instance.A[l,1],self.instance.A[l,2], self.instance.pICTto[l,t].value*self.instance.baseMVA,\
                self.instance.pICTfrom[l,t].value*self.instance.baseMVA])
        print (tabulate(zoneflow, headers="firstrow", tablefmt="grid"))


        print ("***************\n")
        print ("Storage\n")
        print ("***************")
        storage = []
        storage.append(['Storage','Time Period','Charge(MWh)','Discharge(MWh)','State(MWh)'])
        for t in self.instance.T:
            for s in self.instance.S:
                storage.append([s,t,self.instance.pSIn[s,t].value*self.instance.baseMVA,\
                self.instance.pSOut[s,t].value*self.instance.baseMVA,\
                self.instance.pS[s,t].value*self.instance.baseMVA])


        print (tabulate(storage, headers="firstrow", tablefmt="grid"))

        print ("==============================================")

        cols_summary    = ['Time Period','Conventional generation (MW)', 'Wind generation (MW)', 'Demand (MW)']
        cols_zone       = ['Time Period','Zone', 'Conventional generation (MW)', 'Wind generation (MW)', 'Demand (MW)']
        cols_ict        = ['Time Period','From', 'To', 'Power flow To(MW)', 'Power flow Fr(MW)']
        cols_gen        = ['Time Period','Generator', 'on/off', 'start', 'stop','pG(MW)']

        summary         = pd.DataFrame(columns=cols_summary)
        zone            = pd.DataFrame(columns=cols_zone)
        interconnect    = pd.DataFrame(columns=cols_ict)
        generation      = pd.DataFrame(columns=cols_gen)
        ind = 0
        for t in self.instance.T:
            summary.loc[ind] = pd.Series({'Time Period':t,'Conventional generation (MW)':sum(self.instance.pG[g,t].value for g in self.instance.G)*self.instance.baseMVA,\
            'Wind generation (MW)':sum(self.instance.pW[w,t].value for w in self.instance.WIND)*self.instance.baseMVA,\
            'Demand (MW)':sum(self.instance.PD[d,t] for d in self.instance.D)*self.instance.baseMVA})
            ind += 1
        ind = 0
        for t in self.instance.T:
            for z in self.instance.Z:
                zone.loc[ind] = pd.Series({'Time Period':t,'Zone':z,'Conventional generation (MW)':sum(self.instance.pG[g,t].value for (g,z) in self.instance.GZ)*self.instance.baseMVA,\
                'Wind generation (MW)':sum(self.instance.pW[w,t].value for (w,z) in self.instance.WZ)*self.instance.baseMVA,\
                'Demand (MW)':sum(self.instance.PD[d,t] for (d,z) in self.instance.DZ)*self.instance.baseMVA})
                ind += 1
        ind = 0
        for t in self.instance.T:
            for l in self.instance.ICT:
                interconnect.loc[ind] = pd.Series({'Time Period':t,'From':self.instance.A[l,1],'To':self.instance.A[l,2],\
                'Power flow To(MW)':self.instance.pICTto[l,t].value*self.instance.baseMVA,'Power flow Fr(MW)':self.instance.pICTfrom[l,t].value*self.instance.baseMVA})
                ind += 1
        ind = 0
        for t in self.instance.T:
            for g in self.instance.G:
                generation.loc[ind] = pd.Series({'Time Period':t,'Generator':g, 'on/off':self.instance.u[g,t].value,\
                'start':self.instance.ustart[g,t].value, 'stop':self.instance.ustop[g,t].value,\
                'pG(MW)':self.instance.pG[g,t].value*self.instance.baseMVA})
                ind += 1
        #===write output on xlsx file===
        #
        writer = pd.ExcelWriter('results.xlsx', engine ='xlsxwriter')
        summary.to_excel(writer, sheet_name = 'summary',index=False)
        zone.to_excel(writer, sheet_name = 'zone',index=False)
        interconnect.to_excel(writer, sheet_name = 'interconnection',index=False)
        generation.to_excel(writer, sheet_name = 'generator',index=False)
        writer.save()



#==================================================================
# ACOPF.mod
# PYOMO model file of "AC" optimal power flow problem (ACOPF)
# ---Author---
# W. Bukhsh,
# wbukhsh@gmail.com
# OATS
# Copyright (c) 2015 by W Bukhsh, Glasgow, Scotland
# OATS is distributed under the GNU GENERAL PUBLIC LICENSE license. (see LICENSE file for details).
#==========Import==========
from __future__ import division
from pyomo.environ import *
#==========================

model = AbstractModel()
# --- sets ---
# buses, generators, loads, lines, sections
model.B      = Set()  # set of buses
model.G      = Set()  # set of generators
model.WIND   = Set()  # set of wind generators
model.D      = Set()  # set of demands
model.L      = Set()  # set of lines
model.SHUNT  = Set()  # set of shunts
model.LE     = Set()  # line-to and from ends set (1,2)
model.TRANSF = Set()  # set of transformers
model.b0     = Set(within=model.B)  # set of reference buses

# generators, buses, loads linked to each bus b
model.Gbs = Set(within=model.B * model.G)    # generator-bus mapping
model.Dbs = Set(within=model.B * model.D)    # demand-bus mapping
model.Wbs = Set(within=model.B * model.WIND) # wind-bus mapping
model.SHUNTbs = Set(within=model.B * model.SHUNT)# shunt-bus mapping

# --- parameters ---
# line matrix
model.A = Param(model.L*model.LE)       # bus-line
model.AT = Param(model.TRANSF*model.LE) # bus-transformer

# demands
model.PD = Param(model.D, within=Reals)  # real power demand
model.QD = Param(model.D, within=Reals)  # reactive power demand
model.VOLL    = Param(model.D, within=Reals) #value of lost load

# generators
model.PGmax    = Param(model.G, within=NonNegativeReals) # max real power of generator
model.PGmin    = Param(model.G, within=Reals)            # min real power of generator
model.QGmax    = Param(model.G, within=NonNegativeReals) # max reactive power of generator
model.QGmin    = Param(model.G, within=Reals)            # min reactive power of generator

#wind generators
model.WGmax    = Param(model.WIND,  within=NonNegativeReals) # max real power of wind generator
model.WGmin    = Param(model.WIND,  within=NonNegativeReals) # min real power of wind generator
model.WGQmax   = Param(model.WIND,  within=NonNegativeReals) # max reactive power of wind generator
model.WGQmin   = Param(model.WIND,  within=Reals)            # min reactive power of wind generator

# lines
model.SLmax = Param(model.L, within=NonNegativeReals) # max real power limit on flow in a line
model.GL = Param(model.L, within=Reals)
model.BL = Param(model.L, within=Reals)
model.BC = Param(model.L, within=Reals)

#emergency ratings
model.SLmax_E = Param(model.L, within=NonNegativeReals)       # max emergency real power flow limit
model.SLmaxT_E = Param(model.TRANSF, within=NonNegativeReals) # max emergency real power flow limit

#transformers
model.Tap          = Param(model.TRANSF, within=NonNegativeReals)  # turns ratio of a transformer
model.TapLB        = Param(model.TRANSF, within=NonNegativeReals)  # lower bound on turns ratio
model.TapUB        = Param(model.TRANSF, within=NonNegativeReals)  # upper bound on turns ratio
model.Deltashift   = Param(model.TRANSF) #  phase shift of transformer, rad
model.DeltashiftLB = Param(model.TRANSF) #  lower bound on phase shift of transformer, rad
model.DeltashiftUB = Param(model.TRANSF) #  upper bound on phase shift of transformer, rad

model.SLmaxT = Param(model.TRANSF, within=NonNegativeReals) # max real power flow limit
model.GLT    = Param(model.TRANSF, within=Reals)
model.BLT    = Param(model.TRANSF, within=Reals)

# derived line parameters
model.G11 = Param(model.L, within=Reals)
model.G12 = Param(model.L, within=Reals)
model.G21 = Param(model.L, within=Reals)
model.G22 = Param(model.L, within=Reals)
model.B11 = Param(model.L, within=Reals)
model.B12 = Param(model.L, within=Reals)
model.B21 = Param(model.L, within=Reals)
model.B22 = Param(model.L, within=Reals)
## derived transformer parameters
model.G11T = Param(model.TRANSF, within=Reals)
model.G12T = Param(model.TRANSF, within=Reals)
model.G21T = Param(model.TRANSF, within=Reals)
model.G22T = Param(model.TRANSF, within=Reals)
model.B11T = Param(model.TRANSF, within=Reals)
model.B12T = Param(model.TRANSF, within=Reals)
model.B21T = Param(model.TRANSF, within=Reals)
model.B22T = Param(model.TRANSF, within=Reals)

# buses
model.Vmax = Param(model.B, within=NonNegativeReals) #  max voltage angle
model.Vmin = Param(model.B, within=NonNegativeReals) #  min voltage angle

#shunt
model.GB = Param(model.SHUNT, within=Reals) #  shunt conductance
model.BB = Param(model.SHUNT, within=Reals) #  shunt susceptance

# cost
model.c2 = Param(model.G, within=NonNegativeReals)# generator cost coefficient c2 (*pG^2)
model.c1 = Param(model.G, within=NonNegativeReals)# generator cost coefficient c1 (*pG)
model.c0 = Param(model.G, within=NonNegativeReals)# generator cost coefficient c0

model.baseMVA = Param(within=NonNegativeReals)# base MVA

#constants
model.eps = Param(within=NonNegativeReals)

# --- variables ---
model.pG       = Var(model.G, domain= NonNegativeReals)# real power output of generator
model.qG       = Var(model.G, domain= Reals)# reactive power output of generator
model.pW       = Var(model.WIND, domain= Reals) #real power generation from wind
model.qW       = Var(model.WIND, domain= Reals) #reactive power generation from wind
model.pD       = Var(model.D, domain= Reals)# real power absorbed by demand
model.qD       = Var(model.D, domain= Reals)# reactive power absorbed by demand
model.pLfrom   = Var(model.L, domain= Reals) # real power injected at b onto line
model.pLto     = Var(model.L, domain= Reals) # real power injected at b' onto line
model.qLfrom   = Var(model.L, domain= Reals) # reactive power injected at b onto line
model.qLto     = Var(model.L, domain= Reals) # reactive power injected at b' onto line
model.pLfromT  = Var(model.TRANSF, domain= Reals) # real power injected at b onto transformer
model.pLtoT    = Var(model.TRANSF, domain= Reals) # real power injected at b' onto transformer
model.qLfromT  = Var(model.TRANSF, domain= Reals) # reactive power injected at b onto transformer
model.qLtoT    = Var(model.TRANSF, domain= Reals) # reactive power injected at b' onto transformer

#model.deltaL = Var(model.L, domain= Reals) # angle difference across lines
model.delta  = Var(model.B, domain= Reals, initialize=0.0) # voltage phase angle at bus b, rad
model.v      = Var(model.B, domain= NonNegativeReals, initialize=1.0) # voltage magnitude at bus b, rad
model.alpha  = Var(model.D, initialize=1.0, domain= NonNegativeReals)# proportion to supply of load d

# --- cost function ---
def objective(model):
    obj = sum(model.c2[g]*(model.baseMVA*model.pG[g])**2+model.c1[g]*model.baseMVA*model.pG[g]+ model.c0[g] for g in model.G)+\
    sum(model.VOLL[d]*(1-model.alpha[d])*model.baseMVA*model.PD[d] for d in model.D)
    return obj
model.OBJ = Objective(rule=objective, sense=minimize)

# --- Kirchoff's current law at each bus b ---
def KCL_real_def(model, b):
    return sum(model.pG[g] for g in model.G if (b,g) in model.Gbs) +\
    sum(model.pW[w] for w in model.WIND if (b,w) in model.Wbs)==\
    sum(model.pD[d] for d in model.D if (b,d) in model.Dbs)+\
    sum(model.pLfrom[l] for l in model.L if model.A[l,1]==b)+ \
    sum(model.pLto[l] for l in model.L if model.A[l,2]==b)+\
    sum(model.pLfromT[l] for l in model.TRANSF if model.AT[l,1]==b)+ \
    sum(model.pLtoT[l] for l in model.TRANSF if model.AT[l,2]==b)+\
    sum(model.GB[s]*model.v[b]**2 for s in model.SHUNT if (b,s) in model.SHUNTbs)
def KCL_reactive_def(model, b):
    return sum(model.qG[g] for g in model.G if (b,g) in model.Gbs) +\
    sum(model.qW[w] for w in model.WIND if (b,w) in model.Wbs)== \
    sum(model.qD[d] for d in model.D if (b,d) in model.Dbs)+\
    sum(model.qLfrom[l] for l in model.L if model.A[l,1]==b)+ \
    sum(model.qLto[l] for l in model.L if model.A[l,2]==b)+\
    sum(model.qLfromT[l] for l in model.TRANSF if model.AT[l,1]==b)+ \
    sum(model.qLtoT[l] for l in model.TRANSF if model.AT[l,2]==b)-\
    sum(model.BB[s]*model.v[b]**2 for s in model.SHUNT if (b,s) in model.SHUNTbs)
model.KCL_real     = Constraint(model.B, rule=KCL_real_def)
model.KCL_reactive = Constraint(model.B, rule=KCL_reactive_def)

# --- Kirchoff's voltage law on each line ---
def KVL_real_fromend(model,l):
    return model.pLfrom[l] == model.G11[l]*(model.v[model.A[l,1]]**2)+\
    model.v[model.A[l,1]]*model.v[model.A[l,2]]*(model.B12[l]*sin(model.delta[model.A[l,1]]-\
    model.delta[model.A[l,2]])+model.G12[l]*cos(model.delta[model.A[l,1]]-model.delta[model.A[l,2]]))
def KVL_real_toend(model,l):
    return model.pLto[l] == model.G22[l]*(model.v[model.A[l,2]]**2)+\
    model.v[model.A[l,1]]*model.v[model.A[l,2]]*(model.B21[l]*sin(model.delta[model.A[l,2]]-\
    model.delta[model.A[l,1]])+model.G21[l]*cos(model.delta[model.A[l,2]]-model.delta[model.A[l,1]]))
def KVL_reactive_fromend(model,l):
    return model.qLfrom[l] == -model.B11[l]*(model.v[model.A[l,1]]**2)+\
    model.v[model.A[l,1]]*model.v[model.A[l,2]]*(model.G12[l]*sin(model.delta[model.A[l,1]]-\
    model.delta[model.A[l,2]])-model.B12[l]*cos(model.delta[model.A[l,1]]-model.delta[model.A[l,2]]))
def KVL_reactive_toend(model,l):
    return model.qLto[l] == (-model.B22[l]*(model.v[model.A[l,2]]**2)+\
    model.v[model.A[l,1]]*model.v[model.A[l,2]]*(model.G21[l]*sin(model.delta[model.A[l,2]]-\
    model.delta[model.A[l,1]])-model.B21[l]*cos(model.delta[model.A[l,2]]-model.delta[model.A[l,1]])))
model.KVL_real_from     = Constraint(model.L, rule=KVL_real_fromend)
model.KVL_real_to       = Constraint(model.L, rule=KVL_real_toend)
model.KVL_reactive_from = Constraint(model.L, rule=KVL_reactive_fromend)
model.KVL_reactive_to   = Constraint(model.L, rule=KVL_reactive_toend)

# --- Kirchoff's voltage law on each transformer line ---
def KVL_real_fromendTransf(model,l):
    return model.pLfromT[l] == model.G11T[l]*(model.v[model.AT[l,1]]**2)+\
    model.v[model.AT[l,1]]*model.v[model.AT[l,2]]*(model.B12T[l]*sin(model.delta[model.AT[l,1]]-\
    model.delta[model.AT[l,2]])+model.G12T[l]*cos(model.delta[model.AT[l,1]]-model.delta[model.AT[l,2]]))
def KVL_real_toendTransf(model,l):
    return model.pLtoT[l] == model.G22T[l]*(model.v[model.AT[l,2]]**2)+\
    model.v[model.AT[l,1]]*model.v[model.AT[l,2]]*(model.B21T[l]*sin(model.delta[model.AT[l,2]]-\
    model.delta[model.AT[l,1]])+model.G21T[l]*cos(model.delta[model.AT[l,2]]-model.delta[model.AT[l,1]]))
def KVL_reactive_fromendTransf(model,l):
    return model.qLfromT[l] == -model.B11T[l]*(model.v[model.AT[l,1]]**2)+\
    model.v[model.AT[l,1]]*model.v[model.AT[l,2]]*(model.G12T[l]*sin(model.delta[model.AT[l,1]]-\
    model.delta[model.AT[l,2]])-model.B12T[l]*cos(model.delta[model.AT[l,1]]-model.delta[model.AT[l,2]]))
def KVL_reactive_toendTransf(model,l):
    return model.qLtoT[l] == -model.B22T[l]*(model.v[model.AT[l,2]]**2)+\
    model.v[model.AT[l,1]]*model.v[model.AT[l,2]]*(model.G21T[l]*sin(model.delta[model.AT[l,2]]-\
    model.delta[model.AT[l,1]])-model.B21T[l]*cos(model.delta[model.AT[l,2]]-model.delta[model.AT[l,1]]))
model.KVL_real_fromTransf     = Constraint(model.TRANSF, rule=KVL_real_fromendTransf)
model.KVL_real_toTransf       = Constraint(model.TRANSF, rule=KVL_real_toendTransf)
model.KVL_reactive_fromTransf = Constraint(model.TRANSF, rule=KVL_reactive_fromendTransf)
model.KVL_reactive_toTransf   = Constraint(model.TRANSF, rule=KVL_reactive_toendTransf)

# --- generator power limits ---
def Real_Power_Max(model,g):
    return model.pG[g] <= model.PGmax[g]
def Real_Power_Min(model,g):
    return model.pG[g] >= model.PGmin[g]
def Reactive_Power_Max(model,g):
    return model.qG[g] <= model.QGmax[g]
def Reactive_Power_Min(model,g):
    return model.qG[g] >= model.QGmin[g]
model.PGmaxC = Constraint(model.G, rule=Real_Power_Max)
model.PGminC = Constraint(model.G, rule=Real_Power_Min)
model.QGmaxC = Constraint(model.G, rule=Reactive_Power_Max)
model.QGminC = Constraint(model.G, rule=Reactive_Power_Min)

# ---wind generator power limits ---
def Wind_Real_Power_Max(model,w):
    return model.pW[w] <= model.WGmax[w]
def Wind_Real_Power_Min(model,w,t):
    return model.pW[w] >= model.WGmin[w]
def Wind_Reactive_Power_Max(model,w):
    return model.qW[w] <= model.WGQmax[w]
def Wind_Reactive_Power_Min(model,w):
    return model.qW[w] >= model.WGQmin[w]
model.WGmaxC  = Constraint(model.WIND, rule=Wind_Real_Power_Max)
model.WGminC  = Constraint(model.WIND, rule=Wind_Real_Power_Min)
model.WGQmaxC = Constraint(model.WIND, rule=Wind_Reactive_Power_Max)
model.WGQminC = Constraint(model.WIND, rule=Wind_Reactive_Power_Min)

# --- demand and load shedding ---
def Load_Shed_real(model,d):
    return model.pD[d] == model.alpha[d]*model.PD[d]
def Load_Shed_reactive(model,d):
    return model.qD[d] == model.alpha[d]*model.QD[d]
model.LoadShed_real = Constraint(model.D, rule=Load_Shed_real)
model.LoadShed_reactive = Constraint(model.D, rule=Load_Shed_reactive)

def alpha_BoundUB(model,d):
    return model.alpha[d] <= 1
def alpha_BoundLB(model,d):
    return model.alpha[d] >= 0
model.alphaBoundUBC = Constraint(model.D, rule=alpha_BoundUB)
model.alphaBoundLBC = Constraint(model.D, rule=alpha_BoundLB)

# --- line power limits ---
def line_lim1_def(model,l):
    return model.pLfrom[l]**2+model.qLfrom[l]**2 <= model.SLmax[l]**2
def line_lim2_def(model,l):
    return model.pLto[l]**2+model.qLto[l]**2 <= model.SLmax[l]**2
model.line_lim1 = Constraint(model.L, rule=line_lim1_def)
model.line_lim2 = Constraint(model.L, rule=line_lim2_def)

# --- power flow limits on transformer lines---
def transf_lim1_def(model,l):
    return model.pLfromT[l]**2+model.qLfromT[l]**2 <= model.SLmaxT[l]**2
def transf_lim2_def(model,l):
    return model.pLtoT[l]**2+model.qLtoT[l]**2 <= model.SLmaxT[l]**2
model.transf_lim1 = Constraint(model.TRANSF, rule=transf_lim1_def)
model.transf_lim2 = Constraint(model.TRANSF, rule=transf_lim2_def)

# --- voltage constraints ---
def bus_max_voltage(model,b):
    return model.v[b] <= model.Vmax[b]
def bus_min_voltage(model,b):
    return model.v[b] >= model.Vmin[b]
model.Vmaxc = Constraint(model.B, rule=bus_max_voltage)
model.Vminc = Constraint(model.B, rule=bus_min_voltage)

# --- reference bus constraint ---
def ref_bus_def(model,b):
    return model.delta[b]==0
model.refbus = Constraint(model.b0, rule=ref_bus_def)



#==================================================================
# UC.py
# PYOMO model file of unit commitment problem
# ---Author---
# W. Bukhsh,
# wbukhsh@gmail.com
# OATS
# Copyright (c) 2015 by W. Bukhsh, Glasgow, Scotland
# OATS is distributed under the GNU GENERAL PUBLIC LICENSE v3. (see LICENSE file for details).
#==================================================================

#==========Import==========
from __future__ import division
from pyomo.environ import *
#==========================

model = AbstractModel()

# --- SETS ---
model.Z      = Set()  # set of zones
model.G      = Set()  # set of generators
model.WIND   = Set()  # set of wind generators
model.D      = Set()  # set of loads
model.T      = Set()  # set of time periods, as a list of positive integers
model.ICT    = Set()  # set of interconnections between zone
model.TRed   = Set()  # set of time periods T={1,...,N-1} ignoring last time period-req for ramp
model.SHUNT  = Set()  # set of shunts
model.LE     = Set()  # line-to and from ends set (1,2)
model.S      = Set()  # set of storage

## mapping sets
model.GZ  = Set(within=model.G*model.Z)     # set of generators and zones
model.WZ  = Set(within=model.WIND*model.Z)  # set of WIND abd zones
model.DZ  = Set(within=model.D*model.Z)     # set of demand and zones
model.SZ  = Set(within=model.SHUNT*model.Z) # set of shunts and zones
model.Sbs = Set(within=model.Z*model.S)    # set of storage-zone mapping

# UC Min Up and Min Down indexing sets
model.UCMinDownT   = Set(within=model.G*model.T*model.T) #indexing set for min down constraints
model.UCMinUpT     = Set(within=model.G*model.T*model.T) #indexing set for min up constraints
model.MinDownTime  = Param(model.G)
model.MinUpTime    = Param(model.G)

# --- parameters ---

# storage
model.StoreUB       = Param(model.S, within=NonNegativeReals)# max real power capacity of storage, p.u.
model.StoreLB       = Param(model.S, within=NonNegativeReals)# max real power capacity of storage, p.u.
model.StoreInitial  = Param(model.S, within=NonNegativeReals)# charging efficieny of storage
model.StoreFinal    = Param(model.S, within=NonNegativeReals)# charging efficieny of storage
model.ChargeEff     = Param(model.S, within=NonNegativeReals)# charging efficieny of storage
model.DischargeEff  = Param(model.S, within=NonNegativeReals)# charging efficieny of storage
model.RateCharge    = Param(model.S, within=NonNegativeReals)# rate of charging of storage
model.RateDischarge = Param(model.S, within=NonNegativeReals)# rate of charging of storage
model.rampCharge    = Param(model.S, within=NonNegativeReals)# storage cost for discharging
model.rampDischarge = Param(model.S, within=NonNegativeReals)# storage cost for discharging

# interconnector matrix
model.A = Param(model.ICT*model.LE)  # bus-line (node-arc) matrix
# demands
model.PD    = Param(model.D,model.T, within=Reals)  #real power demand
model.VOLL  = Param(model.D, within=Reals)          # Value of lost load
# generators
model.PGmax = Param(model.G, within=NonNegativeReals)# max real power of generator, p.u.
model.PGmin = Param(model.G, within=Reals)# min real power of generator, p.u.

#RES
model.WGmax = Param(model.WIND,model.T, within=NonNegativeReals)# max real power of wind generator, p.u.
model.WGmin = Param(model.WIND,model.T, within=NonNegativeReals)# min real power of wind generator, p.u.

# inter-connector
model.NTCto = Param(model.ICT, within=NonNegativeReals) # to NTC
model.NTCfr = Param(model.ICT, within=NonNegativeReals) # from NTC

#ramping rates
model.RampUp   = Param(model.G, within=NonNegativeReals)
model.RampDown = Param(model.G, within=NonNegativeReals)

# cost
model.c2 = Param(model.G, within=NonNegativeReals)# generator cost coefficient c2 (*pG^2)
model.c1 = Param(model.G, within=NonNegativeReals)# generator cost coefficient c1 (*pG)
model.c0 = Param(model.G, within=NonNegativeReals)# generator cost coefficient c0
model.SUcosts = Param(model.G, within=NonNegativeReals)# generator start-up cost
model.SDcosts = Param(model.G, within=NonNegativeReals)# generator shut-down cost

model.baseMVA = Param(within=NonNegativeReals)# base MVA

#constants
model.eps  = Param(within=NonNegativeReals)


#reserve requirements
model.ResReq = Param(model.Z, within=NonNegativeReals)

model.nT     = Param(within=NonNegativeReals) #total time horizon

#params and sets for boundary conditions
model.UCBoundaryConditions = Param(model.G,model.T) #boundary conditions for the UC problem
model.UCBCGenSet     = Set(within=model.G)  # set of reference buses
model.UCBCTimeSet    = Set(within=model.T)  # set of reference buses

# --- variables ---
model.pG       = Var(model.G,model.T, domain= NonNegativeReals)# real power output of generator g
model.pGResv   = Var(model.G,model.T, domain= NonNegativeReals)# real power reserve from the generator g
model.pW       = Var(model.WIND,model.T, domain= NonNegativeReals)# real power output of RES r
model.pD       = Var(model.D,model.T, domain= Reals)# real power demand delivered at d
model.alpha    = Var(model.D,model.T, domain= NonNegativeReals)# propotion of real power demand delivered at d
model.pICTto   = Var(model.ICT,model.T, domain= Reals) # real power injected at b' onto line l
model.pICTfrom = Var(model.ICT,model.T, domain= Reals) # reactive power injected at b onto line l

#storage variables
model.pS    = Var(model.S,model.T, domain= Reals)  #real power contribution of storage
model.pSIn  = Var(model.S,model.T, domain= NonNegativeReals)  #real power charging
model.pSOut = Var(model.S,model.T, domain= NonNegativeReals)  #real power discharging


#Unit comittment variables
model.u      = Var(model.G,model.T, within=Binary, initialize=1) # unit comittment variables for the generators
model.ustart = Var(model.G,model.T, within=Binary, initialize=1) # unit comittment variables for the generators
model.ustop  = Var(model.G,model.T, within=Binary, initialize=1) # unit comittment variables for the generators

#cost function
model.costfunc  = Var(model.T)

# --- cost function ---
def objective(model):
    obj = sum(model.costfunc[t] for t in model.T)
    return obj
model.OBJ = Objective(rule=objective, sense=minimize)

def cost_def(model, t):
    return model.costfunc[t] == sum((model.c2[g]*(model.pG[g,t])**2+model.c1[g]*model.pG[g,t]+\
    model.c0[g]*model.u[g,t])*model.baseMVA +\
    model.ustart[g,t]*model.SUcosts[g]+\
    model.ustop[g,t]*model.SDcosts[g] for g in model.G)+\
    sum(model.VOLL[d]*(1-model.alpha[d,t])*model.baseMVA*model.PD[d,t] for d in model.D)

# the next line creates one KCL constraint for each bus
model.cost_const = Constraint(model.T, rule=cost_def)

# --- Kirchoff's current law Definition at each bus b ---
def KCL_def(model,z, t):
    return sum(model.pG[g,t] for g in model.G if (g,z) in model.GZ)+\
    sum(model.pW[w,t] for w in model.WIND if (w,z) in model.WZ) +\
    sum(model.pSOut[s,t] for s in model.S if (z,s) in model.Sbs) - \
    sum(model.pSIn[s,t] for s in model.S if (z,s) in model.Sbs)+\
    sum(model.pICTfrom[i,t] for i in model.ICT if model.A[i,1]==z) +\
    sum(model.pICTto[i,t] for i in model.ICT if model.A[i,2]==z) == \
    sum(model.pD[d,t] for d in model.D if (d,z) in model.DZ)+\
    sum(model.GB[s] for s in model.SHUNT if (s,z) in model.SZ)

# the next line creates one KCL constraint for each bus
model.KCL_const = Constraint(model.Z, model.T, rule=KCL_def)

# --- generator power limits ---
def Real_Power_Max(model,g,t):
    return model.pG[g,t]+model.pGResv[g,t] <= model.PGmax[g]*model.u[g,t]
def Real_Power_Min(model,g,t):
    return model.pG[g,t] >= model.PGmin[g]*model.u[g,t]

model.PGmaxC = Constraint(model.G,model.T, rule=Real_Power_Max)
model.PGminC = Constraint(model.G,model.T, rule=Real_Power_Min)

# --- Wind power limits ---
def Wind_Real_Power_Max(model,w,t):
    return model.pW[w,t] <= model.WGmax[w,t]
def Wind_Real_Power_Min(model,w,t):
    return model.pW[w,t] >= model.WGmin[w,t]

model.WindPGmaxC = Constraint(model.WIND,model.T, rule=Wind_Real_Power_Max)
model.WindPGminC = Constraint(model.WIND,model.T, rule=Wind_Real_Power_Min)

# --- Reserve requirements ---
def Reserve_Req(model,z,t):
    return sum(model.pGResv[g,t] for g in model.G if (g,z) in model.GZ) >= model.ResReq[z]
model.ReserveReqC = Constraint(model.Z,model.T, rule=Reserve_Req)

# --- demands and load shedding ---
def Load_Shed_real(model,d,t):
    return model.pD[d,t] == model.alpha[d,t]*model.PD[d,t]

model.LoadShed_real = Constraint(model.D,model.T, rule=Load_Shed_real)

def alpha_BoundUB(model,d,t):
    return model.alpha[d,t] <= 1.0
def alpha_BoundLB(model,d,t):
    return model.alpha[d,t] >= 0.0

model.alphaBoundUBC = Constraint(model.D,model.T, rule=alpha_BoundUB)
model.alphaBoundLBC = Constraint(model.D,model.T, rule=alpha_BoundLB)

# --- generator ramp limits ---
def Real_Power_RampUp(model,g,t):
    return model.pG[g,t]-model.pG[g,t-1] <= model.RampUp[g]
def Real_Power_RampDown(model,g,t):
    return model.pG[g,t]-model.pG[g,t-1] >= -model.RampDown[g]

model.RampUpC   = Constraint(model.G,model.TRed, rule=Real_Power_RampUp)
model.RampDownC = Constraint(model.G,model.TRed, rule=Real_Power_RampDown)

# --- UC constraints ---
def UC_start_shut_cons(model,g,t):
    return model.ustart[g,t]-model.ustop[g,t]== model.u[g,t]-model.u[g,t-1]
model.UC_start_shunt_def = Constraint(model.G,model.TRed, rule=UC_start_shut_cons)

#---min up and down time constraints---
def UC_Min_UpTime_cons(model,g,t1,t2):
    return sum(model.ustart[g,k] for k in range(t1,t2))  <= model.u[g,t2]
def UC_Min_DownTime_cons(model,g,t1,t2):
    return sum(model.ustop[g,k] for k in range(t1,t2))   <= 1-model.u[g,t2]
model.UC_Min_UpTime_def = Constraint(model.UCMinUpT, rule=UC_Min_UpTime_cons)
model.UC_Min_DownTime_def = Constraint(model.UCMinDownT, rule=UC_Min_DownTime_cons)

# --- UC Boundary constraints ---
def UC_BC_cons(model,g,t):
    return model.u[g,t]== model.UCBoundaryConditions[g,t]

model.UC_BC_def = Constraint(model.UCBCGenSet,model.UCBCTimeSet, rule=UC_BC_cons)

# --- line power limits ---
def ICT_lim_def1(model,i,t):
    return model.pICTto[i,t] <= model.NTCto[i]
def ICT_lim_def2(model,i,t):
    return model.pICTfrom[i,t] <= model.NTCfr[i]
def ICT_lim_def3(model,i,t):
    return model.pICTto[i,t] >= -model.NTCto[i]
def ICT_lim_def4(model,i,t):
    return model.pICTfrom[i,t] >= -model.NTCfr[i]
model.line_lim1 = Constraint(model.ICT,model.T, rule=ICT_lim_def1)
model.line_lim2 = Constraint(model.ICT,model.T, rule=ICT_lim_def2)
model.line_lim4 = Constraint(model.ICT,model.T, rule=ICT_lim_def4)
model.line_lim3 = Constraint(model.ICT,model.T, rule=ICT_lim_def3)
# --- Loss equations ---
def Loss_Approx(model,i,t):
    return model.pICTto[i,t]+model.pICTfrom[i,t] == 0

model.Loss_ApproxC = Constraint(model.ICT,model.T, rule=Loss_Approx)

# --- storage model ---
def storage_model(model,s,t):
    return model.pS[s,t] == (model.pSIn[s,t]*(model.ChargeEff[s])-model.pSOut[s,t]*(1/model.DischargeEff[s]))
#the next two lines creates constraints for demand model
model.StoreModelC = Constraint(model.S,model.T, rule=storage_model)


#===Storage dynamics Constraints
def storage_dynamics_UB(model,s,t):
    return sum(model.pS[s,i] for i in range(1,t+1))  <= model.StoreUB[s]-model.StoreInitial[s]
def storage_dynamics_LB(model,s,t):
    return sum(model.pS[s,i] for i in range(1,t+1))  >= model.StoreLB[s]-model.StoreInitial[s]
#the next two lines creates constraints for demand model
model.StoreModelDynUBC = Constraint(model.S,model.TRed, rule=storage_dynamics_UB)
model.StoreModelDynLBC = Constraint(model.S,model.TRed, rule=storage_dynamics_LB)

def storage_dynamics_UB_firtsperiod(model,s):
    return model.pS[s,0]   <= model.StoreUB[s]-model.StoreInitial[s]
def storage_dynamics_LB_firtsperiod(model,s):
    return model.pS[s,0]  >= model.StoreLB[s]-model.StoreInitial[s]
#the next two lines creates constraints for demand model
model.StoreModelDynUBC_fixed = Constraint(model.S,rule=storage_dynamics_UB_firtsperiod)
model.StoreModelDynLBC_fixed = Constraint(model.S,rule=storage_dynamics_LB_firtsperiod)

# --- boundary constraint for storage ---
def storage_boundary_constraint(model,s):
    return sum(model.pS[s,t] for t in model.T) == model.StoreFinal[s]-model.StoreInitial[s]
model.storageBounddef = Constraint(model.S,rule=storage_boundary_constraint)



#==================================================================
# runfile.py
# This is a top level OATS script. Simulations can be ran using this script
# ---Author---
# W. Bukhsh,
# wbukhsh@gmail.com
# OATS
# Copyright (c) 2017 by W. Bukhsh, Glasgow, Scotland
# OATS is distributed under the GNU GENERAL PUBLIC LICENSE v3. (see LICENSE file for details).
#==================================================================
import logging
import os
from oats.runcase import runcase

logging.basicConfig(level=logging.INFO,
                    format='%(asctime)s %(levelname)-8s %(message)s',
                    datefmt='%a, %d %b %Y %H:%M:%S',
                    filename='oatslog.log',
                    filemode='w')
logging.info("OATS log file")
logging.info("Program started")
oats_dir = os.path.dirname(os.path.realpath(__file__))
default_testcase = oats_dir+'/testcases/case24_ieee_rts.xlsx'

#----------------------------------------------------------------------
# DC Load flow
def dclf(tc='default',solver='ipopt',neos=True,out=0):
    """
    Solves DC load flow problem

    ARGUMENTS:
        **tc** (*.xlsx file)  - OATS test case. See OATS data format for details

        **solver** (str)  - name of a solver. Defualt is 'ipopt'

        **neos** (bool) - If True, the problem is solved using NEOS otherwise using a localy install solver.

        **out** (bool) - If True, the output is displayed on screen.
    """

    if tc == 'default':
        tc = default_testcase
    #options
    opt=({'neos':neos,\
    'solver':solver,'out':out})
    testcase = tc
    model ='DCLF'
    # ==log==
    logging.info("Solver selected: "+opt['solver'])
    logging.info("Testcase selected: "+testcase)
    logging.info("Model selected: "+model)
    runcase(testcase,model,opt)
    logging.info("Done!")
# AC Load flow
def aclf(tc='default',solver='ipopt',neos=True,out=0):
    """
    Solves AC load flow problem

    ARGUMENTS:
        **tc** (*.xlsx file)  - OATS test case. See OATS data format for details

        **solver** (str)  - name of a solver. Defualt is 'ipopt'

        **neos** (bool) - If True, the problem is solved using NEOS otherwise using a localy install solver.

        **out** (bool) - If True, the output is displayed on screen.
    """
    if tc == 'default':
        tc = default_testcase
    opt=({'neos':neos,\
    'solver':solver,'out':out})
    testcase = tc
    model ='ACLF'
    # ==log==
    logging.info("Solver selected: "+opt['solver'])
    logging.info("Testcase selected: "+testcase)
    logging.info("Model selected: "+model)
    runcase(testcase,model,opt)
    logging.info("Done!")
# DC optimal power flow problem
def dcopf(tc='default',solver='ipopt',neos=True,out=0):
    """
    Solves DC optimal power flow problem

    ARGUMENTS:
        **tc** (*.xlsx file)  - OATS test case. See OATS data format for details

        **solver** (str)  - name of a solver. Defualt is 'ipopt'

        **neos** (bool) - If True, the problem is solved using NEOS otherwise using a localy install solver.

        **out** (bool) - If True, the output is displayed on screen.
    """

    if tc == 'default':
        tc = default_testcase
    #options
    opt=({'neos':neos,\
    'solver':solver,'out':out})
    testcase = tc
    model ='DCOPF'
    # ==log==
    logging.info("Solver selected: "+opt['solver'])
    logging.info("Testcase selected: "+testcase)
    logging.info("Model selected: "+model)
    runcase(testcase,model,opt)
    logging.info("Done!")

# AC optimal power flow problem
def acopf(tc='default',solver='ipopt',neos=True,out=0):
    """
    Solves AC optimal power flow problem

    ARGUMENTS:
        **tc** (*.xlsx file)  - OATS test case. See OATS data format for details

        **solver** (str)  - name of a solver. Defualt is 'ipopt'

        **neos** (bool) - If True, the problem is solved using NEOS otherwise using a localy install solver.

        **out** (bool) - If True, the output is displayed on screen.
    """

    if tc == 'default':
        tc = default_testcase
    #options
    opt=({'neos':neos,\
    'solver':solver,'out':out})
    testcase = tc
    model ='ACOPF'
    # ==log==
    logging.info("Solver selected: "+opt['solver'])
    logging.info("Testcase selected: "+testcase)
    logging.info("Model selected: "+model)
    runcase(testcase,model,opt)
    logging.info("Done!")

# security constrained optimal power flow problem
def scopf(tc='default',solver='ipopt',neos=True,out=0):
    """
    Solves security constrained optimal power flow problem

    ARGUMENTS:
        **tc** (*.xlsx file)  - OATS test case. See OATS data format for details

        **solver** (str)  - name of a solver. Defualt is 'ipopt'

        **neos** (bool) - If True, the problem is solved using NEOS otherwise using a localy install solver.

        **out** (bool) - If True, the output is displayed on screen.
    """

    if tc == 'default':
        tc = default_testcase
    #options
    opt=({'neos':neos,\
    'solver':solver,'out':out})
    testcase = tc
    model ='SCOPF'
    # ==log==
    logging.info("Solver selected: "+opt['solver'])
    logging.info("Testcase selected: "+testcase)
    logging.info("Model selected: "+model)
    runcase(testcase,model,opt)
    logging.info("Done!")

# unit commitment problem
def uc(tc='default',solver='cplex',neos=True,out=0):
    """
    Solves unit commitment problem

    ARGUMENTS:
        **tc** (*.xlsx file)  - OATS test case. See OATS data format for details

        **solver** (str)  - name of a solver. Defualt is 'ipopt'

        **neos** (bool) - If True, the problem is solved using NEOS otherwise using a localy install solver.

        **out** (bool) - If True, the output is displayed on screen.
    """

    if tc == 'default':
        tc = default_testcase
    #options
    opt=({'neos':neos,\
    'solver':solver,'out':out})
    testcase = tc
    model ='UC'
    # ==log==
    logging.info("Solver selected: "+opt['solver'])
    logging.info("Testcase selected: "+testcase)
    logging.info("Model selected: "+model)
    runcase(testcase,model,opt)
    logging.info("Done!")


# user defined model
def model(model,tc='default',solver='cplex',neos=True,out=0):
    """
    Solves a user defined model

    ARGUMENTS:
        **model** (*.py file)  - A model file that describes the optimisation problem using PYOMO modelling language.

        **tc** (*.xlsx file)  - OATS test case. See OATS data format for details

        **solver** (str)  - name of a solver. Defualt is 'ipopt'

        **neos** (bool) - If True, the problem is solved using NEOS otherwise using a localy install solver.

        **out** (bool) - If True, the output is displayed on screen.
    """

    if tc == 'default':
        tc = default_testcase
    #options
    opt=({'neos':neos,\
    'solver':solver,'out':out,'user_def_model':True})
    testcase = tc
    model =model
    # ==log==
    logging.info("Solver selected: "+opt['solver'])
    logging.info("Testcase selected: "+testcase)
    logging.info("Model selected: "+model)
    runcase(testcase,model,opt)
    logging.info("Done!")



#==================================================================
# ACLF.mod
# PYOMO model file of "AC" load flow problem (ACLF)
# ---Author---
# W. Bukhsh,
# wbukhsh@gmail.com
# OATS
# Copyright (c) 2015 by W Bukhsh, Glasgow, Scotland
# OATS is distributed under the GNU GENERAL PUBLIC LICENSE v3. (see LICENSE file for details).
#==========Import==========
from __future__ import division
from pyomo.environ import *
#==========================

model = AbstractModel()
# --- sets ---
model.B         = Set()  # set of buses
model.G         = Set()  # set of generators
model.WIND      = Set()  # set of wind generators
model.D         = Set()  # set of loads
model.L         = Set()  # set of lines
model.SHUNT     = Set()  # set of shunts
model.LE        = Set()  # line-to and from ends set (1,2)
model.TRANSF    = Set()  # set of transformers
model.Bvolt     = Set()  # set of bus that connects voltage regulated transformer
model.b0        = Set(within=model.B)  # set of reference buses
model.g0        = Set(within=model.G)  # set of generators attached to reference bus
model.Transf2   = Set(within=model.TRANSF)
model.DistSlack = Set(within=model.G)
# generators, buses, loads linked to each bus b
model.Gbs  = Set(within=model.B * model.G)    # bus-generator map
model.Dbs  = Set(within=model.B * model.D)    # bus-demand map
model.Wbs  = Set(within=model.B * model.WIND) # bus-wind map
model.SHUNTbs = Set(within=model.B * model.SHUNT) # bus-shunt map

# --- parameters ---
model.A = Param(model.L*model.LE)        # bus-line
model.AT = Param(model.TRANSF*model.LE)  # bus-transformer

# demands
model.PD = Param(model.D, within=Reals)  # real power demand
model.QD = Param(model.D, within=Reals)  # real power demand
model.VOLL    = Param(model.D, within=Reals) #value of lost load

model.PG  = Param(model.G, within=NonNegativeReals)    # real power set-point of generator
model.WG  = Param(model.WIND, within=NonNegativeReals) # real power set-point of wind generator

# line chracteristics
model.BC = Param(model.L, within=Reals)

#transformers
model.Deltashift   = Param(model.TRANSF) #  phase shift of transformer, rad
model.GLT    = Param(model.TRANSF, within=Reals)
model.BLT    = Param(model.TRANSF, within=Reals)

# derived line parameters
model.G11 = Param(model.L, within=Reals)
model.G12 = Param(model.L, within=Reals)
model.G21 = Param(model.L, within=Reals)
model.G22 = Param(model.L, within=Reals)
model.B11 = Param(model.L, within=Reals)
model.B12 = Param(model.L, within=Reals)
model.B21 = Param(model.L, within=Reals)
model.B22 = Param(model.L, within=Reals)
## derived transformer parameters
model.G11T = Param(model.TRANSF, within=Reals)
model.G12T = Param(model.TRANSF, within=Reals)
model.G21T = Param(model.TRANSF, within=Reals)
model.G22T = Param(model.TRANSF, within=Reals)
model.B11T = Param(model.TRANSF, within=Reals)
model.B12T = Param(model.TRANSF, within=Reals)
model.B21T = Param(model.TRANSF, within=Reals)
model.B22T = Param(model.TRANSF, within=Reals)

model.g  = Param(model.TRANSF, within=Reals)
model.b  = Param(model.TRANSF, within=Reals)
model.bC = Param(model.TRANSF, within=Reals)

# --- tap ratio bounds ---
model.TapUB = Param(model.TRANSF, within=NonNegativeReals)
model.TapLB = Param(model.TRANSF, within=NonNegativeReals)
model.Tap = Param(model.TRANSF, within=NonNegativeReals)
model.VTar = Param(model.Bvolt, within=NonNegativeReals)
# voltage targets
model.VS = Param(model.G, within=NonNegativeReals) #  voltage target at generator buses

##shunt
model.GB = Param(model.SHUNT, within=Reals) #  shunt conductance
model.BB = Param(model.SHUNT, within=Reals) #  shunt susceptance

model.baseMVA = Param(within=NonNegativeReals)# base MVA

#constants
model.eps = Param(within=NonNegativeReals)

# --- variables ---
model.pG       = Var(model.G, domain= NonNegativeReals)# real power output of generator
model.qG       = Var(model.G, domain= Reals)    # reactive power output of generator
model.pW       = Var(model.WIND, domain= Reals) #real power generation from wind
model.qW       = Var(model.WIND, domain= Reals) #reactive power generation from wind
model.pD       = Var(model.D, domain= Reals) # real power absorbed by demand
model.qD       = Var(model.D, domain= Reals) # reactive power absorbed by demand
model.pLfrom   = Var(model.L, domain= Reals) # real power injected at b onto line
model.pLto     = Var(model.L, domain= Reals) # real power injected at b' onto line
model.qLfrom   = Var(model.L, domain= Reals) # reactive power injected at b onto line
model.qLto     = Var(model.L, domain= Reals) # reactive power injected at b' onto line
model.pLfromT  = Var(model.TRANSF, domain= Reals) # real power injected at b onto transformer
model.pLtoT    = Var(model.TRANSF, domain= Reals) # real power injected at b' onto transformer
model.qLfromT  = Var(model.TRANSF, domain= Reals) # reactive power injected at b onto transformer
model.qLtoT    = Var(model.TRANSF, domain= Reals) # reactive power injected at b' onto transformer
model.delta    = Var(model.B, domain= Reals, initialize=0.0) # voltage phase angle at bus b, rad
model.v        = Var(model.B, domain= NonNegativeReals, initialize=1.0) # voltage magnitude at bus b, rad
model.tap      = Var(model.TRANSF,initialize=1.0, within=NonNegativeReals)  # turns ratio of a transformer

# variables that model the violations (to help with the convergence)
model.epsPG_up    = Var(model.DistSlack, initialize=0.0, domain= NonNegativeReals)
model.epsPG_down  = Var(model.DistSlack, initialize=0.0, domain= NonNegativeReals)
model.epsV_up     = Var(model.G, initialize=0.0, domain= NonNegativeReals)
model.epsV_down   = Var(model.G, initialize=0.0, domain= NonNegativeReals)
# --- cost function ---
def objective(model):
    obj = 1*sum(model.epsV_up[g]+model.epsV_down[g] for g in model.G)+\
    1*sum(model.baseMVA*(model.pLto[l]+model.pLfrom[l]) for l in model.L)+\
    sum(model.baseMVA*(model.epsPG_up[g]+model.epsPG_down[g]) for g in model.DistSlack)
    return obj
model.OBJ = Objective(rule=objective, sense=minimize)

# --- Kirchoff's current law at each bus b ---
def KCL_real_def(model, b):
    return sum(model.pG[g] for g in model.G if (b,g) in model.Gbs) +\
    sum(model.pW[w] for w in model.WIND if (b,w) in model.Wbs)==\
    sum(model.PD[d] for d in model.D if (b,d) in model.Dbs)+\
    sum(model.pLfrom[l] for l in model.L if model.A[l,1]==b)+ \
    sum(model.pLto[l] for l in model.L if model.A[l,2]==b)+\
    sum(model.pLfromT[l] for l in model.TRANSF if model.AT[l,1]==b)+ \
    sum(model.pLtoT[l] for l in model.TRANSF if model.AT[l,2]==b)+\
    sum(model.GB[s]*model.v[b]**2 for s in model.SHUNT if (b,s) in model.SHUNTbs)
def KCL_reactive_def(model, b):
    return sum(model.qG[g] for g in model.G if (b,g) in model.Gbs) +\
    sum(model.qW[w] for w in model.WIND if (b,w) in model.Wbs)== \
    sum(model.QD[d] for d in model.D if (b,d) in model.Dbs)+\
    sum(model.qLfrom[l] for l in model.L if model.A[l,1]==b)+ \
    sum(model.qLto[l] for l in model.L if model.A[l,2]==b)+\
    sum(model.qLfromT[l] for l in model.TRANSF if model.AT[l,1]==b)+ \
    sum(model.qLtoT[l] for l in model.TRANSF if model.AT[l,2]==b)-\
    sum(model.BB[s]*model.v[b]**2 for s in model.SHUNT if (b,s) in model.SHUNTbs)
model.KCL_real     = Constraint(model.B, rule=KCL_real_def)
model.KCL_reactive = Constraint(model.B, rule=KCL_reactive_def)

# --- Kirchoff's voltage law on each line ---
def KVL_real_fromend(model,l):
    return model.pLfrom[l] == (model.G11[l]*(model.v[model.A[l,1]]**2)+\
    model.v[model.A[l,1]]*model.v[model.A[l,2]]*(model.B12[l]*sin(model.delta[model.A[l,1]]-\
    model.delta[model.A[l,2]])+model.G12[l]*cos(model.delta[model.A[l,1]]-model.delta[model.A[l,2]])))
def KVL_real_toend(model,l):
    return model.pLto[l] == (model.G22[l]*(model.v[model.A[l,2]]**2)+\
    model.v[model.A[l,1]]*model.v[model.A[l,2]]*(model.B21[l]*sin(model.delta[model.A[l,2]]-\
    model.delta[model.A[l,1]])+model.G21[l]*cos(model.delta[model.A[l,2]]-model.delta[model.A[l,1]])))
def KVL_reactive_fromend(model,l):
    return model.qLfrom[l] == (-model.B11[l]*(model.v[model.A[l,1]]**2)+\
    model.v[model.A[l,1]]*model.v[model.A[l,2]]*(model.G12[l]*sin(model.delta[model.A[l,1]]-\
    model.delta[model.A[l,2]])-model.B12[l]*cos(model.delta[model.A[l,1]]-model.delta[model.A[l,2]])))
def KVL_reactive_toend(model,l):
    return model.qLto[l] == (-model.B22[l]*(model.v[model.A[l,2]]**2)+\
    model.v[model.A[l,1]]*model.v[model.A[l,2]]*(model.G21[l]*sin(model.delta[model.A[l,2]]-\
    model.delta[model.A[l,1]])-model.B21[l]*cos(model.delta[model.A[l,2]]-model.delta[model.A[l,1]])))
model.KVL_real_from     = Constraint(model.L, rule=KVL_real_fromend)
model.KVL_real_to       = Constraint(model.L, rule=KVL_real_toend)
model.KVL_reactive_from = Constraint(model.L, rule=KVL_reactive_fromend)
model.KVL_reactive_to   = Constraint(model.L, rule=KVL_reactive_toend)

# --- Kirchoff's voltage law on each transformer line ---
def KVL_real_fromendTransf(model,l):
    if (model.AT[l,1] in model.Bvolt) or (model.AT[l,1] in model.Bvolt):
        return model.pLfromT[l] == (model.g[l]/(model.tap[l]**2))*(model.v[model.AT[l,1]]**2)-\
        (1/model.tap[l])*model.v[model.AT[l,1]]*model.v[model.AT[l,2]]*(model.b[l]*sin(model.delta[model.AT[l,1]]-\
        model.delta[model.AT[l,2]])+model.g[l]*cos(model.delta[model.AT[l,1]]-model.delta[model.AT[l,2]]))
    else:
        return model.pLfromT[l] == (model.g[l]/(model.Tap[l]**2))*(model.v[model.AT[l,1]]**2)-\
        (1/model.Tap[l])*model.v[model.AT[l,1]]*model.v[model.AT[l,2]]*(model.b[l]*sin(model.delta[model.AT[l,1]]-\
        model.delta[model.AT[l,2]])+model.g[l]*cos(model.delta[model.AT[l,1]]-model.delta[model.AT[l,2]]))
def KVL_real_toendTransf(model,l):
    if (model.AT[l,1] in model.Bvolt) or (model.AT[l,1] in model.Bvolt):
        return model.pLtoT[l] == model.g[l]*(model.v[model.AT[l,2]]**2)-\
        (1/model.tap[l])*model.v[model.AT[l,1]]*model.v[model.AT[l,2]]*(model.b[l]*sin(model.delta[model.AT[l,2]]-\
        model.delta[model.AT[l,1]])+model.g[l]*cos(model.delta[model.AT[l,2]]-model.delta[model.AT[l,1]]))
    else:
        return model.pLtoT[l] == model.g[l]*(model.v[model.AT[l,2]]**2)-\
        (1/model.Tap[l])*model.v[model.AT[l,1]]*model.v[model.AT[l,2]]*(model.b[l]*sin(model.delta[model.AT[l,2]]-\
        model.delta[model.AT[l,1]])+model.g[l]*cos(model.delta[model.AT[l,2]]-model.delta[model.AT[l,1]]))
def KVL_reactive_fromendTransf(model,l):
    if (model.AT[l,1] in model.Bvolt) or (model.AT[l,1] in model.Bvolt):
        return model.qLfromT[l] == -(model.b[l]+0.5*model.bC[l])*(model.v[model.AT[l,1]]**2)/(model.tap[l]**2)-\
        (1/model.tap[l])*model.v[model.AT[l,1]]*model.v[model.AT[l,2]]*(model.g[l]*sin(model.delta[model.AT[l,1]]-\
        model.delta[model.AT[l,2]])-model.b[l]*cos(model.delta[model.AT[l,1]]-model.delta[model.AT[l,2]]))
    else:
        return model.qLfromT[l] == -(model.b[l]+0.5*model.bC[l])*(model.v[model.AT[l,1]]**2)/(model.Tap[l]**2)-\
        (1/model.Tap[l])*model.v[model.AT[l,1]]*model.v[model.AT[l,2]]*(model.g[l]*sin(model.delta[model.AT[l,1]]-\
        model.delta[model.AT[l,2]])-model.b[l]*cos(model.delta[model.AT[l,1]]-model.delta[model.AT[l,2]]))
def KVL_reactive_toendTransf(model,l):
    if (model.AT[l,1] in model.Bvolt) or (model.AT[l,1] in model.Bvolt):
        return model.qLtoT[l] == -(model.b[l]+0.5*model.bC[l])*(model.v[model.AT[l,2]]**2)-\
        (1/model.tap[l])*model.v[model.AT[l,1]]*model.v[model.AT[l,2]]*(model.g[l]*sin(model.delta[model.AT[l,2]]-\
        model.delta[model.AT[l,1]])-model.b[l]*cos(model.delta[model.AT[l,2]]-model.delta[model.AT[l,1]]))
    else:
        return model.qLtoT[l] == -(model.b[l])*(model.v[model.AT[l,2]]**2)-\
        (1/model.Tap[l])*model.v[model.AT[l,1]]*model.v[model.AT[l,2]]*(model.g[l]*sin(model.delta[model.AT[l,2]]-\
        model.delta[model.AT[l,1]])-model.b[l]*cos(model.delta[model.AT[l,2]]-model.delta[model.AT[l,1]]))
model.KVL_real_fromTransf     = Constraint(model.TRANSF, rule=KVL_real_fromendTransf)
model.KVL_real_toTransf       = Constraint(model.TRANSF, rule=KVL_real_toendTransf)
model.KVL_reactive_fromTransf = Constraint(model.TRANSF, rule=KVL_reactive_fromendTransf)
model.KVL_reactive_toTransf   = Constraint(model.TRANSF, rule=KVL_reactive_toendTransf)

# --- generator power limits ---
def Real_Power_up(model,g):
    if g in model.DistSlack:
        return model.pG[g] <= model.PG[g]+model.epsPG_up[g]
    else:
        return model.pG[g] <= model.PG[g]
def Real_Power_down(model,g):
    if g in model.DistSlack:
        return model.pG[g] >= model.PG[g]-model.epsPG_down[g]
    else:
        return model.pG[g] >= model.PG[g]
model.PGmaxC = Constraint(model.G-model.g0, rule=Real_Power_up)
model.PGminC = Constraint(model.G-model.g0, rule=Real_Power_down)

# ---wind generator power limits ---
def Wind_Real_Power_up(model,w):
    return model.pW[w] <= model.WG[w]
def Wind_Real_Power_down(model,w):
    return model.pW[w] >= model.WG[w]
model.WGmaxC  = Constraint(model.WIND, rule=Wind_Real_Power_up)
model.WGminC  = Constraint(model.WIND, rule=Wind_Real_Power_down)


# --- voltage target constraints ---
def bus_max_voltage(model,b,g):
    return model.v[b] <= model.VS[g]+model.epsV_up[g]
def bus_min_voltage(model,b,g):
    return model.v[b] >= model.VS[g]-model.epsV_down[g]
model.Vmaxc = Constraint(model.Gbs, rule=bus_max_voltage)
model.Vminc = Constraint(model.Gbs, rule=bus_min_voltage)

# --- tap ratios ---
def tap_max(model,l):
    if l in model.Transf2:
        return model.tap[l] <= model.TapUB[l]
    else:
        return model.tap[l] <= model.Tap[l]
def tap_min(model,l):
    if l in model.Transf2:
        return model.tap[l] >= model.TapLB[l]
    else:
        return model.tap[l] >= model.Tap[l]
model.TapMaxC = Constraint(model.TRANSF, rule=tap_max)
model.TapMinC = Constraint(model.TRANSF, rule=tap_min)

# --- voltage constraint on a voltage regulated transformer---
def volt_reg(model,b):
    return model.v[b] <= model.VTar[b]
def volt_reg2(model,b):
    return model.v[b] >= model.VTar[b]
model.VoltRegC  = Constraint(model.Bvolt, rule=volt_reg)
model.VoltRegC2 = Constraint(model.Bvolt, rule=volt_reg2)

# --- reference bus constraint ---
def ref_bus_def(model,b):
    return model.delta[b]==0
model.refbus = Constraint(model.b0, rule=ref_bus_def)



#==================================================================
# DCLF.mod
# PYOMO model file of "DC" load flow problem (DCLF)
# This formulation uses the standard "DC" linearization of the AC power flow equations
# ---Author---
# W. Bukhsh,
# wbukhsh@gmail.com
# OATS
# Copyright (c) 2015 by W Bukhsh, Glasgow, Scotland
# OATS is distributed under the GNU GENERAL PUBLIC LICENSE v3. (see LICENSE file for details).
#==================================================================

#==========Import==========
from __future__ import division
from pyomo.environ import *
#==========================

model = AbstractModel()

# --- SETS ---
model.B      = Set()  # set of buses
model.G      = Set()  # set of generators
model.WIND   = Set()  # set of wind generators
model.D      = Set()  # set of demands
model.L      = Set()  # set of lines
model.SHUNT  = Set()  # set of shunts
model.b0     = Set(within=model.B)  # set of reference buses
model.LE     = Set()  # line-to and from ends set (1,2)
model.TRANSF = Set()  # set of transformers

# generators, buses, loads linked to each bus b
model.Gbs     = Set(within=model.B * model.G)    # set of generator-bus mapping
model.Wbs     = Set(within=model.B * model.WIND) # set of wind-bus mapping
model.Dbs     = Set(within=model.B * model.D)    # set of demand-bus mapping
model.SHUNTbs = Set(within=model.B*model.SHUNT)  # set of shunt-bus mapping

# --- parameters ---
# line matrix
model.A     = Param(model.L*model.LE)       # bus-line (node-arc) matrix
model.AT    = Param(model.TRANSF*model.LE)  # bus-transformer (node-arc) matrix

# demands
model.PD      = Param(model.D, within=Reals)  # real power demand
model.VOLL    = Param(model.D, within=Reals)  # value of lost load

# generators
model.PG    = Param(model.G, within=Reals)    # real power set-point of generator
model.WG    = Param(model.WIND, within=NonNegativeReals) # real power set-point of wind generator

# lines and transformer chracteristics
model.BL  = Param(model.L, within=Reals)        # susceptance of a line
model.BLT = Param(model.TRANSF, within=Reals)  # susceptance of a transformer
model.Tap = Param(model.TRANSF, within=NonNegativeReals)  # turns ratio of transformer

# shunt
model.GB = Param(model.SHUNT, within=Reals) #  shunt conductance
model.BB = Param(model.SHUNT, within=Reals) #  shunt susceptance

model.baseMVA = Param(within=NonNegativeReals)# base MVA

#constants
model.eps = Param(within=NonNegativeReals)


# --- control variables ---
model.pG    = Var(model.G, domain= Reals) #real power generation
model.pW    = Var(model.WIND, domain= Reals)         #real power generation from wind
model.pD    = Var(model.D, domain= Reals)            #real power demand delivered
model.alpha = Var(model.D, domain= NonNegativeReals) #propotion of real power demand delivered
# --- state variables ---
model.deltaL  = Var(model.L, domain= Reals)      # angle difference across lines
model.deltaLT = Var(model.TRANSF, domain= Reals) # angle difference across transformers
model.delta   = Var(model.B, domain= Reals, initialize=0.0) # voltage phase angle at bus b, rad
model.pL      = Var(model.L, domain= Reals)      # real power injected at b onto line l
model.pLT     = Var(model.TRANSF, domain= Reals) # real power injected at b onto transformer line l
# --- variables that model small violations on PG values. this is to help with the convergence
model.epsG_up      = Var(model.G, domain= NonNegativeReals)
model.epsG_down    = Var(model.G, domain= NonNegativeReals)
model.epsW_up      = Var(model.WIND, domain= NonNegativeReals)
model.epsW_down    = Var(model.WIND, domain= NonNegativeReals)

# --- cost function ---
def objective(model):
    obj = sum(model.baseMVA*(model.epsG_down[g]+model.epsG_up[g]) for (b,g) in model.Gbs if b not in model.b0)+\
    sum(model.baseMVA*(model.epsW_down[w]+model.epsW_up[w]) for w in model.WIND)+\
    sum(model.baseMVA*model.VOLL[d]*(1-model.alpha[d])*model.PD[d] for d in model.D)
    return obj
model.OBJ = Objective(rule=objective, sense=minimize)

# --- Kirchoff's current law at each bus b ---
def KCL_def(model, b):
    return sum(model.pG[g] for g in model.G if (b,g) in model.Gbs) +\
    sum(model.pW[w] for w in model.WIND if (b,w) in model.Wbs) == \
    sum(model.pD[d] for d in model.D if (b,d) in model.Dbs)+\
    sum(model.pL[l] for l in model.L if model.A[l,1]==b)- \
    sum(model.pL[l] for l in model.L if model.A[l,2]==b)+\
    sum(model.pLT[l] for l in model.TRANSF if model.AT[l,1]==b)- \
    sum(model.pLT[l] for l in model.TRANSF if model.AT[l,2]==b)+\
    sum(model.GB[s] for s in model.SHUNT if (b,s) in model.SHUNTbs)
model.KCL_const = Constraint(model.B, rule=KCL_def)

# --- Kirchoff's voltage law at each line and transformer---
def KVL_line_def(model,l):
    return model.pL[l] == (-model.BL[l])*model.deltaL[l]
def KVL_trans_def(model,l):
    return model.pLT[l] == (-model.BLT[l]/model.Tap[l])*model.deltaLT[l]
model.KVL_line_const     = Constraint(model.L, rule=KVL_line_def)
model.KVL_trans_const    = Constraint(model.TRANSF, rule=KVL_trans_def)

# --- demand model ---
def demand_model(model,d):
    return model.pD[d] == model.alpha[d]*model.PD[d]
def demand_LS_bound_Max(model,d):
    return model.alpha[d] <= 1
model.demandmodelC = Constraint(model.D, rule=demand_model)
model.demandalphaC = Constraint(model.D, rule=demand_LS_bound_Max)

# --- generator power limits ---
def Real_Power_Max(model,g):
    return model.pG[g] <= model.PG[g]+model.epsG_up[g]
def Real_Power_Min(model,g):
    return model.pG[g] >= model.PG[g]-model.epsG_down[g]
model.PGmaxC = Constraint(model.G, rule=Real_Power_Max)
model.PGminC = Constraint(model.G, rule=Real_Power_Min)

# ---wind generator power limits ---
def Wind_Real_Power_Max(model,w):
    return model.pW[w] <= model.WG[w]+model.epsW_up[w]
def Wind_Real_Power_Min(model,w):
    return model.pW[w] >= model.WG[w]+model.epsW_up[w]
model.WGmaxC = Constraint(model.WIND, rule=Wind_Real_Power_Max)
model.WGminC = Constraint(model.WIND, rule=Wind_Real_Power_Min)

# --- phase angle constraints ---
def phase_angle_diff1(model,l):
    return model.deltaL[l] == model.delta[model.A[l,1]] - \
    model.delta[model.A[l,2]]
model.phase_diff1 = Constraint(model.L, rule=phase_angle_diff1)

# --- reference bus constraint ---
def ref_bus_def(model,b):
    return model.delta[b]==0
model.refbus = Constraint(model.b0, rule=ref_bus_def)



#==================================================================
# runcase.py
# This is a second level OATS script, which is called by the runfile.
# This script receives model,testcase, options as an input to run simulation
# ---Author---
# W. Bukhsh,
# wbukhsh@gmail.com
# April 2018
# OATS
# Copyright (c) 2017 by W. Bukhsh, Glasgow, Scotland
# OATS is distributed under the GNU GENERAL PUBLIC LICENSE v3. (see LICENSE file for details).
#==================================================================

#===============Import===============
from __future__ import division
import os
from pyomo.environ import *
from pyomo.opt import SolverFactory
from pyomo.opt import SolverStatus, TerminationCondition
import logging
from oats.selecttestcase import selecttestcase
from oats.printdata import printdata
from oats.printoutput import printoutput
import imp
#====================================

def runcase(testcase,mod,opt=None):
    oats_dir = os.path.dirname(os.path.realpath(__file__))
    if 'user_def_model' in opt:
        modelf = imp.load_source(mod, mod+'.py')
        model = modelf.model
    else:
        try:
            modelf = imp.load_source(mod, oats_dir+'/models/'+mod+'.py')
            model = modelf.model
            logging.info("Given model file found and selected from the models library")
        except Exception:
            logging.error("Given model file not found in the 'models' library", exc_info=False)
            raise
    try:
        ptc = selecttestcase(testcase) #read test case
        logging.info("Given testcase file found and selected from the testcase library")
    except Exception:
        logging.error("Given testcase  not found in the 'testcases' library", exc_info=False)
        raise
    datfile = 'datafile.dat'
    r = printdata(datfile,ptc,mod,opt)
    r.reducedata()
    r.printheader()
    if 'UC' in mod:
        r.printUCdat()
    else:
        r.printkeysets()
        r.printnetwork()
        #'OPF' or 'LF'
        if 'OPF' in mod:
            r.printOPF()
        elif 'LF' in mod:
            r.printLF()
        #'AC' or 'DC'
        if 'AC' in mod:
            r.printAC()
        elif 'DC' or 'SC' in mod:
            r.printDC()

        if mod=='ACLF':
            r.printACLF()
        elif mod=='DCOPF':
            r.printDCOPF()
        elif 'ACOPF' in mod:
            r.printACOPF()
        elif mod=='SCOPF' or mod=='SCOPF_BM':
            r.printSCdat()
        if mod=='ACOPF_BM':
            r.printBM()
        if mod=='DCOPF_BM' or mod=='SCOPF_BM':
            r.printDCBM()



    ###############Solver settings####################
    if (not opt['neos']):

        optimise = SolverFactory(opt['solver'])
        #opt.options['mipgap'] = 0.1
        #################################################

        ############Solve###################
        instance = model.create_instance(datfile)
        instance.dual = Suffix(direction=Suffix.IMPORT)
        results = optimise.solve(instance,tee=True)
        instance.solutions.load_from(results)
        # ##################################
        #
        # ############Output###################
        o = printoutput(results, instance,mod)
        if (opt['out']):
            o.solutionstatus()
        else:
            o.greet()
            o.solutionstatus()
        if 'UC' in mod:
            o.printUC()
        else:
            o.printsummary()
            o.printoutputxls()

    else:
        instance       = model.create_instance(datfile)
        solveroptions  = SolverFactory(opt['solver'])
        solver_manager = SolverManagerFactory('neos')
        results        = solver_manager.solve(instance, opt=solveroptions)
        print (results)



#==================================================================
# DCOPF.mod
# PYOMO model file of "DC" optimal power flow problem (DCOPF)
# This formulation uses the standard "DC" linearization of the AC power flow equations
# ---Author---
# W. Bukhsh,
# wbukhsh@gmail.com
# OATS
# Copyright (c) 2017 by W Bukhsh, Glasgow, Scotland
# OATS is distributed under the GNU GENERAL PUBLIC LICENSE v3. (see LICENSE file for details).
#==================================================================

#==========Import==========
from __future__ import division
from pyomo.environ import *
#==========================

model = AbstractModel()

# --- SETS ---
model.B      = Set()  # set of buses
model.G      = Set()  # set of generators
model.WIND   = Set()  # set of wind generators
model.D      = Set()  # set of demands
model.L      = Set()  # set of lines
model.SHUNT  = Set()  # set of shunts
model.b0     = Set(within=model.B)  # set of reference buses
model.LE     = Set()  # line-to and from ends set (1,2)
model.TRANSF = Set()  # set of transformers

# generators, buses, loads linked to each bus b
model.Gbs     = Set(within=model.B * model.G)    # set of generator-bus mapping
model.Wbs     = Set(within=model.B * model.WIND) # set of wind-bus mapping
model.Dbs     = Set(within=model.B * model.D)    # set of demand-bus mapping
model.SHUNTbs = Set(within=model.B*model.SHUNT)  # set of shunt-bus mapping

# --- parameters ---
# line matrix
model.A     = Param(model.L*model.LE)       # bus-line matrix
model.AT    = Param(model.TRANSF*model.LE)  # bus-transformer matrix

# demands
model.PD      = Param(model.D, within=Reals)  # real power demand
model.VOLL    = Param(model.D, within=Reals)  # value of lost load
# generators
model.PGmax    = Param(model.G, within=NonNegativeReals) # max real power of generator, p.u.
model.PGmin    = Param(model.G, within=Reals)            # min real power of generator, p.u.
model.PG       = Param(model.G, within=Reals) # FPN
model.WGmax    = Param(model.WIND, within=NonNegativeReals) # max real power of wind generator, p.u.
model.WGmin    = Param(model.WIND, within=NonNegativeReals) # min real power of wind generator, p.u.

# lines and transformer chracteristics and ratings
model.SLmax  = Param(model.L, within=NonNegativeReals)      # real power line limit
model.SLmaxT = Param(model.TRANSF, within=NonNegativeReals) # real power transformer limit
model.BL     = Param(model.L, within=Reals)       # susceptance of a line
model.BLT    = Param(model.TRANSF, within=Reals)  # susceptance of a transformer

# shunt
model.GB = Param(model.SHUNT, within=Reals) #  shunt conductance
model.BB = Param(model.SHUNT, within=Reals) #  shunt susceptance

# cost data
model.c2    = Param(model.G, within=NonNegativeReals)# generator cost coefficient c2 (*pG^2)
model.c1    = Param(model.G, within=NonNegativeReals)# generator cost coefficient c1 (*pG)
model.c0    = Param(model.G, within=NonNegativeReals)# generator cost coefficient c0

model.baseMVA = Param(within=NonNegativeReals)# base MVA

#constants
model.eps = Param(within=NonNegativeReals)

# --- control variables ---
model.pG    = Var(model.G,  domain= Reals)  #real power generation
model.pW    = Var(model.WIND, domain= Reals) #real power generation from wind
model.pD    = Var(model.D, domain= Reals) #real power demand delivered
model.alpha = Var(model.D, domain= NonNegativeReals) #propotion of real power demand delivered
# --- state variables ---
model.deltaL  = Var(model.L, domain= Reals)      # angle difference across lines
model.deltaLT = Var(model.TRANSF, domain= Reals) # angle difference across transformers
model.delta   = Var(model.B, domain= Reals, initialize=0.0) # voltage phase angle at bus b, rad
model.pL      = Var(model.L, domain= Reals) # real power injected at b onto line l, p.u.
model.pLT     = Var(model.TRANSF, domain= Reals) # real power injected at b onto transformer line l, p.u.

# --- cost function ---
def objective(model):
    obj = sum(model.c2[g]*(model.baseMVA*model.pG[g])**2+\
    model.c1[g]*(model.baseMVA*model.pG[g])+model.c0[g] for g in model.G) +\
    sum(model.VOLL[d]*(1-model.alpha[d])*model.baseMVA*model.PD[d] for d in model.D)
    return obj
model.OBJ = Objective(rule=objective, sense=minimize)

# --- Kirchoff's current law at each bus b ---
def KCL_def(model, b):
    return sum(model.pG[g] for g in model.G if (b,g) in model.Gbs) +\
    sum(model.pW[w] for w in model.WIND if (b,w) in model.Wbs) == \
    sum(model.pD[d] for d in model.D if (b,d) in model.Dbs)+\
    sum(model.pL[l] for l in model.L if model.A[l,1]==b)- \
    sum(model.pL[l] for l in model.L if model.A[l,2]==b)+\
    sum(model.pLT[l] for l in model.TRANSF if model.AT[l,1]==b)- \
    sum(model.pLT[l] for l in model.TRANSF if model.AT[l,2]==b)+\
    sum(model.GB[s] for s in model.SHUNT if (b,s) in model.SHUNTbs)
model.KCL_const = Constraint(model.B, rule=KCL_def)

# --- Kirchoff's voltage law on each line and transformer---
def KVL_line_def(model,l):
    return model.pL[l] == (-model.BL[l])*model.deltaL[l]
def KVL_trans_def(model,l):
    return model.pLT[l] == (-model.BLT[l])*model.deltaLT[l]
model.KVL_line_const     = Constraint(model.L, rule=KVL_line_def)
model.KVL_trans_const    = Constraint(model.TRANSF, rule=KVL_trans_def)

# --- demand model ---
def demand_model(model,d):
    return model.pD[d] == model.alpha[d]*model.PD[d]
def demand_LS_bound_Max(model,d):
    return model.alpha[d] <= 1
model.demandmodelC = Constraint(model.D, rule=demand_model)
model.demandalphaC = Constraint(model.D, rule=demand_LS_bound_Max)

# --- generator power limits ---
def Real_Power_Max(model,g):
    return model.pG[g] <= model.PGmax[g]
def Real_Power_Min(model,g):
    return model.pG[g] >= model.PGmin[g]

model.PGmaxC = Constraint(model.G, rule=Real_Power_Max)
model.PGminC = Constraint(model.G, rule=Real_Power_Min)

# ---wind generator power limits ---
def Wind_Real_Power_Max(model,w):
    return model.pW[w] <= model.WGmax[w]
def Wind_Real_Power_Min(model,w):
    return model.pW[w] >= model.WGmin[w]
model.WGmaxC = Constraint(model.WIND, rule=Wind_Real_Power_Max)
model.WGminC = Constraint(model.WIND, rule=Wind_Real_Power_Min)

# --- line power limits ---
def line_lim1_def(model,l):
    return model.pL[l] <= model.SLmax[l]
def line_lim2_def(model,l):
    return model.pL[l] >= -model.SLmax[l]
model.line_lim1 = Constraint(model.L, rule=line_lim1_def)
model.line_lim2 = Constraint(model.L, rule=line_lim2_def)

# --- power flow limits on transformer lines---
def transf_lim1_def(model,l):
    return model.pLT[l] <= model.SLmaxT[l]
def transf_lim2_def(model,l):
    return model.pLT[l] >= -model.SLmaxT[l]
model.transf_lim1 = Constraint(model.TRANSF, rule=transf_lim1_def)
model.transf_lim2 = Constraint(model.TRANSF, rule=transf_lim2_def)

# --- phase angle constraints ---
def phase_angle_diff1(model,l):
    return model.deltaL[l] == model.delta[model.A[l,1]] - \
    model.delta[model.A[l,2]]
model.phase_diff1 = Constraint(model.L, rule=phase_angle_diff1)

# --- reference bus constraint ---
def ref_bus_def(model,b):
    return model.delta[b]==0
model.refbus = Constraint(model.b0, rule=ref_bus_def)
